import os
import json
import numpy as np
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class DatasetFactoryWildMe(DatasetFactory):
    def create_catalogue_wildme(self, prefix: str, year: int) -> pd.DataFrame:
        # Get paths for annotation JSON file and for folder with images
        path_json = os.path.join(prefix + '.coco', 'annotations', 'instances_train' + str(year) + '.json')
        path_images = os.path.join(prefix + '.coco', 'images', 'train' + str(year))

        # Load annotations JSON file
        with open(os.path.join(self.root, path_json)) as file:
            data = json.load(file)

        # Check whether segmentation is different from a box
        for ann in data['annotations']:
            if len(ann['segmentation']) != 1:
                raise(Exception('Wrong number of segmentations'))
        
        # Extract the data from the JSON file
        create_dict = lambda i: {'identity': i['name'], 'bbox': utils.segmentation_bbox(i['segmentation'][0]), 'image_id': i['image_id'], 'category_id': i['category_id'], 'segmentation': i['segmentation'][0], 'orientation': i['viewpoint']}
        df_annotation = pd.DataFrame([create_dict(i) for i in data['annotations']])
        create_dict = lambda i: {'file_name': i['file_name'], 'image_id': i['id'], 'date': i['date_captured']}
        df_images = pd.DataFrame([create_dict(i) for i in data['images']])
        species = pd.DataFrame(data['categories'])
        species = species.rename(columns={'id': 'category_id', 'name': 'species'})

        # Merge the information from the JSON file
        df = pd.merge(df_annotation, species, how='left', on='category_id')
        df = pd.merge(df, df_images, how='left', on='image_id')

        # Modify some columns
        df['path'] = path_images + os.path.sep + df['file_name']
        df['id'] = range(len(df))
        df.loc[df['identity'] == '____', 'identity'] = self.unknown_name

        # Remove segmentations which are the same as bounding boxes
        ii = []
        for i in range(len(df)):
            ii.append(utils.is_annotation_bbox(df.iloc[i]['segmentation'], df.iloc[i]['bbox'], tol=3))
        df.loc[ii, 'segmentation'] = np.nan

        # Rename empty dates
        df.loc[df['date'] == 'NA', 'date'] = np.nan

        # Remove superficial columns
        df = df.drop(['image_id', 'file_name', 'supercategory', 'category_id'], axis=1)
        df.rename({'id': 'image_id'}, axis=1, inplace=True)
        return self.finalize_catalogue(df)



import os
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class CattleMuzzle(DatasetFactory):
    summary = summary['CattleMuzzle']
    url = 'https://cloud.une.edu.au/index.php/s/eMwaHAPK08dCDru/download'
    archive = 'Cattle Identification (supplementary material).zip'

    @classmethod
    def _download(cls):
        utils.download_url(cls.url, cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)
    
    def create_catalogue(self) -> pd.DataFrame:
        # Find all images in root
        data = utils.find_images(self.root)
        folders = data['path'].str.split(os.path.sep, expand=True)
        n_folders = max(folders.columns)

        # Extract information
        image_id = data['file'].apply(lambda x: x.split('_')[1].split('.')[0])
        identity = folders[n_folders]

        # Finalize the dataframe
        df = pd.DataFrame({
            'image_id': image_id,
            'path': data['path'] + os.path.sep + data['file'],
            'identity': identity,
        })
        return self.finalize_catalogue(df)



import os
import shutil
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class SealID(DatasetFactory):
    summary = summary['SealID']
    prefix = 'source_'
    archive = '22b5191e-f24b-4457-93d3-95797c900fc0_ui65zipk.zip'
    
    @classmethod
    def _download(cls, url=None):
        if url is None:
            raise(Exception('URL must be provided for SealID.\nCheck https://wildlifedatasets.github.io/wildlife-datasets/downloads/#sealid'))
        utils.download_url(url, cls.archive)

    @classmethod
    def _extract(cls, **kwargs):
        utils.extract_archive(cls.archive, delete=True)
        utils.extract_archive(os.path.join('SealID', 'full images.zip'), delete=True)
        utils.extract_archive(os.path.join('SealID', 'patches.zip'), delete=True)
        
        # Create new folder for segmented images
        folder_new = os.getcwd() + 'Segmented'
        if not os.path.exists(folder_new):
            os.makedirs(folder_new)
        
        # Move segmented images to new folder
        folder_move = os.path.join('patches', 'segmented')
        shutil.move(folder_move, os.path.join(folder_new, folder_move))
        folder_move = os.path.join('full images', 'segmented_database')
        shutil.move(folder_move, os.path.join(folder_new, folder_move))
        folder_move = os.path.join('full images', 'segmented_query')
        shutil.move(folder_move, os.path.join(folder_new, folder_move))
        file_copy = os.path.join('patches', 'annotation.csv')
        shutil.copy(file_copy, os.path.join(folder_new, file_copy))
        file_copy = os.path.join('full images', 'annotation.csv')
        shutil.copy(file_copy, os.path.join(folder_new, file_copy))            

    def create_catalogue(self) -> pd.DataFrame:
        # Load information about the dataset
        data = pd.read_csv(os.path.join(self.root, 'full images', 'annotation.csv'))

        # Finalize the dataframe
        df = pd.DataFrame({    
            'image_id': data['file'].str.split('.', expand=True)[0],
            'path': 'full images' + os.path.sep + self.prefix + data['reid_split'] + os.path.sep + data['file'],
            'identity': data['class_id'].astype(int),
            'original_split': data['segmentation_split'].replace({'training': 'train', 'testing': 'test', 'validation': 'val'}),
            'original_split_reid': data['reid_split'],
        })
        return self.finalize_catalogue(df)


class SealIDSegmented(SealID):
    prefix = 'segmented_'
    warning = '''You are trying to download or extract a segmented dataset.
        It is already included in its non-segmented version.
        Skipping.'''
    
    @classmethod
    def get_data(cls, *args, **kwargs):
        print(cls.warning)

    @classmethod
    def _download(cls, *args, **kwargs):
        print(cls.warning)

    @classmethod
    def _extract(cls, *args, **kwargs):
        print(cls.warning)



import os
import json
import numpy as np
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class NDD20(DatasetFactory):
    outdated_dataset = True
    summary = summary['NDD20']
    url = 'https://data.ncl.ac.uk/ndownloader/files/22774175'
    archive = 'NDD20.zip'

    @classmethod
    def _download(cls):
        utils.download_url(cls.url, cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)    

    def create_catalogue(self) -> pd.DataFrame:
        # Load information about the above-water dataset
        with open(os.path.join(self.root, 'ABOVE_LABELS.json')) as file:
            data = json.load(file)
        
        # Analyze the information about the above-water dataset
        entries = []
        for key in data.keys():
            regions = data[key]['regions']
            for region in regions:
                if 'id' in region['region_attributes']:
                    identity = region['region_attributes']['id']
                else:
                    identity = self.unknown_name
                segmentation = np.zeros(2*len(region['shape_attributes']['all_points_x']))
                segmentation[0::2] = region['shape_attributes']['all_points_x']
                segmentation[1::2] = region['shape_attributes']['all_points_y']
                entries.append({
                    'identity': identity,
                    'species': region['region_attributes']['species'],
                    'out_of_focus': np.nan,
                    'file_name': data[key]['filename'],
                    'reg_type': region['shape_attributes']['name'],
                    'segmentation': segmentation,
                    'orientation': 'above'
                })
        
        # Load information about the below-water dataset
        with open(os.path.join(self.root, 'BELOW_LABELS.json')) as file:
            data = json.load(file)
            
        # Analyze the information about the below-water dataset
        for key in data.keys():
            regions = data[key]['regions']
            for region in regions:
                if 'id' in region['region_attributes']:
                    identity = region['region_attributes']['id']
                else:
                    identity = self.unknown_name
                segmentation = np.zeros(2*len(region['shape_attributes']['all_points_x']))
                segmentation[0::2] = region['shape_attributes']['all_points_x']
                segmentation[1::2] = region['shape_attributes']['all_points_y']
                entries.append({
                    'identity': identity,
                    'species': 'WBD',
                    'out_of_focus': region['region_attributes']['out of focus'] == 'true',
                    'file_name': data[key]['filename'],
                    'reg_type': region['shape_attributes']['name'],
                    'segmentation': segmentation,
                    'orientation': 'below'
                })

        # Create the dataframe from entries 
        df = pd.DataFrame(entries)
        if len(df.reg_type.unique()) != 1:
            raise(Exception('Multiple segmentation types'))

        # Finalize the dataframe
        df['image_id'] = range(len(df))
        df['path'] = df['orientation'].str.upper() + os.path.sep + df['file_name']
        df = df.drop(['reg_type', 'file_name'], axis=1)
        return self.finalize_catalogue(df)


class NDD20v2(NDD20):
    outdated_dataset = False

    def fix_labels(self, df: pd.DataFrame) -> pd.DataFrame:
        for i, df_row in df.iterrows():
            # Rewrite wrong segmentations. There is no dolphin -> should be deleted.
            # But that would break compability and the identity is unknown anyway.
            if len(df_row['segmentation']) == 4:
                img = utils.load_image(os.path.join(self.root, df_row['path']))
                w, h = img.size
                df.at[i, 'segmentation'] = np.array(utils.bbox_segmentation([0, 0, w, h]))
        return df



from __future__ import annotations
import os
from copy import deepcopy
import pandas as pd
import numpy as np
from typing import Optional, List, Union, Callable, Tuple
import json
from PIL import Image
import matplotlib.pyplot as plt
import pycocotools.mask as mask_coco
from .summary import summary
from . import utils

class WildlifeDataset:
    """Base class for creating datasets.

    Attributes:    
      df (pd.DataFrame): A full dataframe of the data.
      summary (dict): Summary of the dataset.
      root (str): Root directory for the data.
      update_wrong_labels(bool): Whether `fix_labels` should be called.
      unknown_name (str): Name of the unknown class.
      outdated_dataset (bool): Tracks whether dataset was replaced by a new version.
      determined_by_df (bool): Specifies whether dataset is completely determined by its dataframe.
      saved_to_system_folder (bool): Specifies whether dataset is saved to system (hidden) folders.
      remove_columns (bool): Specifies whether constant columns are removed in `finalize_catalogue`.
      check_files (bool): Specifies whether files should be checks for existence in `finalize_catalogue`.
      transform (Callable): Applied transform when loading the image.
      img_load (str): Applied transform when loading the image.
      labels_string (List[str]): List of labels in strings.
      col_path (str): Column name containing image paths.
      col_label (str): Column name containing individual animal names (labels).
    """

    unknown_name = 'unknown'
    outdated_dataset = False
    determined_by_df = True
    saved_to_system_folder = False
    download_warning = '''You are trying to download an already downloaded dataset.
        This message may have happened to due interrupted download or extract.
        To force the download use the `force=True` keyword such as
        get_data(..., force=True) or download(..., force=True).
        '''
    download_mark_name = 'already_downloaded'
    license_file_name = 'LICENSE_link'

    def __init__(
            self, 
            root: Optional[str] = None,
            df: Optional[pd.DataFrame] = None,
            update_wrong_labels: bool = True,
            transform: Optional[Callable] = None,
            img_load: str = "full",
            remove_unknown: bool = False,
            remove_columns: bool = False,
            check_files: bool = True,
            load_label: bool = False,
            col_path: str = "path",
            col_label: str = "identity",            
            **kwargs) -> None:
        """Initializes the class.

        If `df` is specified, it copies it. Otherwise, it creates it
        by the `create_catalogue` method.

        Args:
            root (Optional[str], optional): Root directory for the data.
            df (Optional[pd.DataFrame], optional): A full dataframe of the data.
            update_wrong_labels (bool, optional): Whether `fix_labels` should be called.
            transform (Optional[Callable], optional): Applied transform when loading the image.
            img_load (str, optional): Applied transform when loading the image.
            remove_unknown (bool, optional): Whether unknown identities should be removed.
            remove_columns (bool, optional): Whether constant columns are removed in `finalize_catalogue`.
            check_files (bool, optional): Whether files should be checks for existence in `finalize_catalogue`.
            load_label (bool, optional): Whether dataset[k] should return only image or also identity.
            col_path (str, optional): Column name containing image paths.
            col_label (str, optional): Column name containing individual animal names (labels).
        """
        
        if not self.saved_to_system_folder and not root is None and not os.path.exists(root):
            raise Exception('root does not exist. You may have have mispelled it.')
        if self.outdated_dataset:
            print('This dataset is outdated. You may want to call a newer version such as %sv2.' % self.__class__.__name__)
        self.update_wrong_labels = update_wrong_labels
        self.root = root
        self.col_path = col_path
        self.col_label = col_label
        self.remove_columns = remove_columns
        self.check_files = check_files
        # If df is not provided, create it
        if df is None:
            df = self.create_catalogue(**kwargs)
        else:
            if not self.determined_by_df:
                print('This dataset is not determined by dataframe. But you construct it so.')
        if remove_unknown:
            df = df[df[self.col_label] != self.unknown_name]
        self.df = df.reset_index(drop=True)
        self.transform = transform
        self.img_load = img_load
        if self.img_load == "auto":
            if "segmentation" in self.df:
                self.img_load = "bbox_mask"
            elif "bbox" in self.df:
                self.img_load = "bbox"
            else:
                self.img_load = "full"
        self.load_label = load_label

    @property
    def labels_string(self):
        return self.df[self.col_label].astype(str).to_numpy()

    @property
    def num_classes(self):
        return self.df[self.col_label].nunique()

    @property
    def metadata(self):
        return self.df
        
    @metadata.setter
    def metadata(self, value):
        self.df = value	
        
    def __len__(self):
        return len(self.df)

    def __getitem__(self, idx: int) -> Image:
        """Load an image with iloc `idx` with transforms `self.transform` and `self.img_load` applied.

        Args:
            idx (int): Index of the image.

        Returns:
            Loaded image.
        """

        img = self.get_image(idx)
        img = self.apply_segmentation(img, idx)
        if self.load_label:
            return img, self.df[self.col_label].iloc[idx]
        else:
            return img

    def get_subset(self, idx: Union[List[int], List[bool]]) -> WildlifeDataset:
        """Returns a subset of the class.

        Args:
            idx (Union[List[int], List[bool]]): Indices in the dataframe of the subset.

        Returns:
            The subset class.
        """

        dataset = deepcopy(self)
        if len(self) == len(idx):
            dataset.df = dataset.df[idx].reset_index(drop=True)
        else:
            dataset.df = dataset.df.loc[idx].reset_index(drop=True)
        return dataset

    def get_image(self, idx: int) -> Image:
        """Load an image with iloc `idx`.

        Args:
            idx (int): Index of the image.

        Returns:
            Loaded image.
        """

        data = self.df.iloc[idx]
        if self.root:
            img_path = os.path.join(self.root, data[self.col_path])
        else:
            img_path = data[self.col_path]
        img = self.load_image(img_path)
        return img
    
    def load_image(self, path: str) -> Image:
        """Load an image with `path`.

        Args:
            path (str): Path to the image.

        Returns:
            Loaded image.
        """

        return utils.load_image(path)

    def apply_segmentation(self, img: Image, idx: int) -> Image:
        """Applies segmentation or bounding box when loading an image.

        Args:
            img (Image): Loaded image.
            idx (int): Index of the image.

        Returns:
            Loaded image.
        """

        # Prepare for segmentations        
        if self.img_load in ["full_mask", "full_hide", "bbox_mask", "bbox_hide"]:
            data = self.df.iloc[idx]
            if not ("segmentation" in data):
                raise ValueError(f"{self.img_load} selected but no segmentation found.")
            segmentation = data["segmentation"]
            if isinstance(segmentation, list) or isinstance(segmentation, np.ndarray):
                # Convert polygon to compressed RLE
                w, h = img.size
                rles = mask_coco.frPyObjects([segmentation], h, w)
                segmentation = mask_coco.merge(rles)
            elif isinstance(segmentation, dict) and (isinstance(segmentation['counts'], list) or isinstance(segmentation['counts'], np.ndarray)):            
                # Convert uncompressed RLE to compressed RLE
                h, w = segmentation['size']
                segmentation = mask_coco.frPyObjects(segmentation, h, w)
            elif isinstance(segmentation, str):
                # Load image mask and convert it to compressed RLE
                segmentation = np.asfortranarray(utils.load_image(os.path.join(self.root, segmentation)))
                if segmentation.ndim == 3:
                    segmentation = segmentation[:,:,0]
                segmentation = mask_coco.encode(segmentation)
            elif not np.any(pd.isnull(segmentation)):
                raise Exception('Segmentation type not recognized')
        # Prepare for bounding boxes
        if self.img_load in ["bbox"]:
            data = self.df.iloc[idx]
            if not ("bbox" in data):
                raise ValueError(f"{self.img_load} selected but no bbox found.")
            if type(data["bbox"]) == str:
                bbox = json.loads(data["bbox"])
            else:
                bbox = data["bbox"]
        
        # Load full image as it is.
        if self.img_load == "full":
            img = img
        # Mask background using segmentation mask.
        elif self.img_load == "full_mask":
            if not np.any(pd.isnull(segmentation)):
                mask = mask_coco.decode(segmentation).astype("bool")
                img = Image.fromarray(img * mask[..., np.newaxis])
        # Hide object using segmentation mask
        elif self.img_load == "full_hide":
            if not np.any(pd.isnull(segmentation)):
                mask = mask_coco.decode(segmentation).astype("bool")
                img = Image.fromarray(img * ~mask[..., np.newaxis])
        # Crop to bounding box
        elif self.img_load == "bbox":
            if not np.any(pd.isnull(bbox)):
                img = img.crop((bbox[0], bbox[1], bbox[0] + bbox[2], bbox[1] + bbox[3]))
        # Mask background using segmentation mask and crop to bounding box.
        elif self.img_load == "bbox_mask":
            if (not np.any(pd.isnull(segmentation))):
                mask = mask_coco.decode(segmentation).astype("bool")
                img = Image.fromarray(img * mask[..., np.newaxis])
                img = utils.crop_black(img)
        # Hide object using segmentation mask and crop to bounding box.
        elif self.img_load == "bbox_hide":
            if (not np.any(pd.isnull(segmentation))):
                mask = mask_coco.decode(segmentation).astype("bool")
                img = Image.fromarray(img * ~mask[..., np.newaxis])
                img = utils.crop_black(img)
        # Crop black background around images
        elif self.img_load == "crop_black":
            img = utils.crop_black(img)
        # Crop white background around images
        elif self.img_load == "crop_white":
            img = utils.crop_white(img)
        else:
            raise ValueError(f"Invalid img_load argument: {self.img_load}")

        if self.transform:
            img = self.transform(img)

        return img

    @classmethod
    def get_data(
            cls,
            root: str,
            force: bool = False,
            **kwargs
            ) -> None:
        """Downloads and extracts the data. Wrapper around `cls._download` and `cls._extract.`

        Args:
            root (str): Where the data should be stored.
            force (bool, optional): It the root exists, whether it should be overwritten.
        """

        dataset_name = cls.__name__
        mark_file_name = os.path.join(root, cls.download_mark_name)
        
        already_downloaded = os.path.exists(mark_file_name)
        if not cls.saved_to_system_folder and already_downloaded and not force:
            print('DATASET %s: DOWNLOADING STARTED.' % dataset_name)
            print(cls.download_warning)
        else:
            print('DATASET %s: DOWNLOADING STARTED.' % dataset_name)
            cls.download(root, force=force, **kwargs)
            print('DATASET %s: EXTRACTING STARTED.' % dataset_name)
            cls.extract(root,  **kwargs)
            print('DATASET %s: FINISHED.\n' % dataset_name)

    @classmethod
    def download(
            cls,
            root: str,
            force: bool = False,
            **kwargs
            ) -> None:
        """Downloads the data. Wrapper around `cls._download`.

        Args:
            root (str): Where the data should be stored.
            force (bool, optional): It the root exists, whether it should be overwritten.
        """
        
        dataset_name = cls.__name__
        mark_file_name = os.path.join(root, cls.download_mark_name)
        
        already_downloaded = os.path.exists(mark_file_name)
        if cls.saved_to_system_folder:
            cls._download(**kwargs)
        elif already_downloaded and not force:
            print('DATASET %s: DOWNLOADING STARTED.' % dataset_name)            
            print(cls.download_warning)
        else:
            if os.path.exists(mark_file_name):
                os.remove(mark_file_name)
            with utils.data_directory(root):
                cls._download(**kwargs)
            open(mark_file_name, 'a').close()
            if hasattr(cls, 'summary') and 'licenses_url' in cls.summary:
                with open(os.path.join(root, cls.license_file_name), 'w') as file:
                    file.write(cls.summary['licenses_url'])
        
    @classmethod    
    def extract(cls, root: str, **kwargs) -> None:
        """Extract the data. Wrapper around `cls._extract`.

        Args:
            root (str): Where the data should be stored.
        """

        if cls.saved_to_system_folder:
            cls._extract(**kwargs)
        else:
            with utils.data_directory(root):
                cls._extract(**kwargs)
            mark_file_name = os.path.join(root, cls.download_mark_name)
            open(mark_file_name, 'a').close()
    
    @classmethod
    def display_name(cls) -> str:
        """Returns name of the dataset without the v2 ending.

        Returns:
            Name of the dataset.
        """

        cls_parent = cls.__bases__[0]
        while cls_parent != object and cls_parent.outdated_dataset:
            cls = cls_parent
            cls_parent = cls.__bases__[0]            
        return cls.__name__

    def _download(self):
        """Downloads the dataset. Needs to be implemented by subclasses.

        Raises:
            NotImplementedError: Needs to be implemented by subclasses.
        """

        raise NotImplementedError('Needs to be implemented by subclasses.')

    def _extract(self):
        """Extracts the dataset. Needs to be implemented by subclasses.

        Raises:
            NotImplementedError: Needs to be implemented by subclasses.
        """

        raise NotImplementedError('Needs to be implemented by subclasses.')

    def create_catalogue(self):
        """Creates the dataframe.

        Raises:
            NotImplementedError: Needs to be implemented by subclasses.
        """

        raise NotImplementedError('Needs to be implemented by subclasses.')
    
    def fix_labels(self, df: pd.DataFrame) -> pd.DataFrame:
        """Fixes labels in dataframe.
        
        Automatically called in `finalize_catalogue`.                
        """

        return df

    def fix_labels_replace_identity(
            self,
            df: pd.DataFrame,
            replace_identity: List[Tuple],
            col: str = 'identity'
            ) -> pd.DataFrame:
        """Replaces all instances of identities.

        Args:
            df (pd.DataFrame): A full dataframe of the data.
            replace_identity (List[Tuple]): List of (old_identity, new_identity)
            col (str, optional): Column to replace in.

        Returns:
            A full dataframe of the data.
        """
        for old_identity, new_identity in replace_identity:
            df[col] = df[col].replace({old_identity: new_identity})
        return df

    def fix_labels_remove_identity(
            self,
            df: pd.DataFrame,
            identities_to_remove: List,
            col: str = 'identity'
            ) -> pd.DataFrame:
        """Removes all instances of identities.

        Args:
            df (pd.DataFrame): A full dataframe of the data.
            identities_to_remove (List): List of identities to remove.
            col (str, optional): Column to remove from.

        Returns:
            A full dataframe of the data.
        """
        idx_remove = [identity in identities_to_remove for identity in df[col]]
        return df[~np.array(idx_remove)]

    def fix_labels_replace_images(
            self,
            df: pd.DataFrame,
            replace_identity: List[Tuple],
            col: str = 'identity'
            ) -> pd.DataFrame:
        """Replaces specified images with specified identities.

        It looks for a subset of image_name in df[self.col_path].
        It may cause problems with `os.path.sep`.

        Args:
            df (pd.DataFrame): A full dataframe of the data.
            replace_identity (List[Tuple]): List of (image_name, old_identity, new_identity).
            col (str, optional): Column to replace in.

        Returns:
            A full dataframe of the data.
        """
        for image_name, old_identity, new_identity in replace_identity:
            n_replaced = 0
            for index, df_row in df.iterrows():
                # Check that there is a image with the required name and identity 
                if image_name in df_row[self.col_path] and old_identity == df_row[col]:
                    df.loc[index, col] = new_identity
                    n_replaced += 1
            if n_replaced == 0:
                print('File name %s with identity %s was not found.' % (image_name, str(old_identity)))
            elif n_replaced > 1:
                print('File name %s with identity %s was found multiple times.' % (image_name, str(old_identity)))
        return df

    def finalize_catalogue(
            self,
            df: pd.DataFrame,
            ) -> pd.DataFrame:
        """Reorders the dataframe and check file paths.

        Reorders the columns and removes constant columns.
        Checks if columns are in correct formats.
        Checks if ids are unique and if all files exist.

        Args:
            df (pd.DataFrame): A full dataframe of the data.

        Returns:
            A full dataframe of the data, slightly modified.
        """

        if self.update_wrong_labels:
            df = self.fix_labels(df)
        self.rename_column(df, 'path', self.col_path)
        self.rename_column(df, 'identity', self.col_label)
        self.check_required_columns(df)
        self.check_types_columns(df)
        df = self.reorder_df(df)
        if self.remove_columns:
            df = self.remove_constant_columns(df)
        self.check_unique_id(df)
        if self.check_files:
            self.check_files_exist(df[self.col_path])
            self.check_files_names(df[self.col_path])
            if 'segmentation' in df.columns:
                self.check_files_exist(df['segmentation'])
        return df

    def rename_column(self, df, name_old, name_new):
        if name_old != name_new:
            if name_new in df.columns:
                raise Exception(f'Column {name_old} already present in dataframe. Cannot rename {name_old} to it.')
            else:
                return df.rename({name_old: name_new}, axis=1, inplace=True)

    def check_required_columns(self, df: pd.DataFrame) -> None:
        """Check if all required columns are present.

        Args:
            df (pd.DataFrame): A full dataframe of the data.
        """

        for col_name in ['image_id', self.col_label, self.col_path]:
            if col_name not in df.columns:
                raise(Exception('Column %s must be in the dataframe columns.' % col_name))

    def check_types_columns(self, df: pd.DataFrame) -> None:
        """Checks if columns are in correct formats.

        The format are specified in `requirements`, which is list
        of tuples. The first value is the name of the column
        and the second value is a list of formats. The column
        must be at least one of the formats.

        Args:
            df (pd.DataFrame): A full dataframe of the data.
        """

        requirements = [
            ('image_id', ['int', 'str']),
            (self.col_label, ['int', 'str']),
            (self.col_path, ['str']),
            ('bbox', ['list_numeric']),
            ('date', ['date']),
            ('keypoints', ['list_numeric']),
            ('position', ['str']),
            ('species', ['str', 'list']),
            ('video', ['int']),
        ]
        # Verify if the columns are in correct formats
        for col_name, allowed_types in requirements:
            if col_name in df.columns:
                # Remove empty values to be sure
                col = df[col_name][~df[col_name].isnull()]
                if len(col) > 0:
                    self.check_types_column(col, col_name, allowed_types)
    
    def check_types_column(self, col: pd.Series, col_name: str, allowed_types: List[str]) -> None:
        """Checks if the column `col` is in the format `allowed_types`.

        Args:
            col (pd.Series): Column to be checked.
            col_name (str): Column name used only for raising exceptions.
            allowed_types (List[str]): List of strings with allowed values:
                `int` (all values must be integers),
                `str` (strings),
                `list` (lists),
                `list_numeric` (lists with numeric values),
                `date` (dates as tested by `pd.to_datetime`).
        """

        if 'int' in allowed_types and pd.api.types.is_integer_dtype(col):
            return None
        if 'str' in allowed_types and pd.api.types.is_string_dtype(col):
            return None
        if 'list' in allowed_types and pd.api.types.is_list_like(col):
            check = True
            for val in col:
                if not pd.api.types.is_list_like(val):
                    check = False
                    break
            if check:                
                return None        
        if 'list_numeric' in allowed_types and pd.api.types.is_list_like(col):
            check = True
            for val in col:            
                if not pd.api.types.is_list_like(val) and not pd.api.types.is_numeric_dtype(pd.Series(val)):
                    check = False
                    break
            if check:                
                return None
        if 'date' in allowed_types:
            try:
                pd.to_datetime(col)
                return None
            except:
                pass
        raise(Exception('Column %s has wrong type. Allowed types = %s' % (col_name, str(allowed_types))))

    def reorder_df(self, df: pd.DataFrame) -> pd.DataFrame:
        """Reorders rows and columns in the dataframe.

        Rows are sorted based on id.
        Columns are reorder based on the `default_order` list.

        Args:
            df (pd.DataFrame): A full dataframe of the data.

        Returns:
            A full dataframe of the data, slightly modified.
        """

        default_order = ['image_id', self.col_label, self.col_path, 'bbox', 'date', 'keypoints', 'orientation', 'segmentation', 'species']
        df_names = list(df.columns)
        col_names = []
        for name in default_order:
            if name in df_names:
                col_names.append(name)
        for name in df_names:
            if name not in default_order:
                col_names.append(name)
        
        df = df.sort_values('image_id').reset_index(drop=True)
        return df.reindex(columns=col_names)

    def remove_constant_columns(self, df: pd.DataFrame) -> pd.DataFrame:
        """Removes columns with a single unique value.

        Args:
            df (pd.DataFrame): A full dataframe of the data.

        Returns:
            A full dataframe of the data, slightly modified.
        """ 

        for df_name in list(df.columns):
            if df[df_name].astype('str').nunique() == 1:
                df = df.drop([df_name], axis=1)
        return df

    def check_unique_id(self, df: pd.DataFrame) -> None:
        """Checks if values in the id column are unique.

        Args:
            df (pd.DataFrame): A full dataframe of the data.
        """

        if len(df['image_id'].unique()) != len(df):
            raise(Exception('Image ID not unique.'))

    def check_files_exist(self, col: pd.Series) -> None:
        """Checks if paths in a given column exist.

        Args:
            col (pd.Series): A column of a dataframe.
        """

        for path in col:
            if type(path) == str and not os.path.exists(os.path.join(self.root, path)):
                raise(Exception('Path does not exist:' + os.path.join(self.root, path)))

    def check_files_names(self, col: pd.Series) -> None:
        """Checks if paths contain .

        Args:
            col (pd.Series): A column of a dataframe.
        """

        for path in col:
            try:
                path.encode("iso-8859-1")
            except UnicodeEncodeError:
                raise(Exception('Characters in path may cause problems. Please use only ISO-8859-1 characters: ' + os.path.join(path)))

    def plot_grid(
            self,
            n_rows: int = 5,
            n_cols: int = 8,
            offset: float = 10,
            img_min: float = 100,
            rotate: bool = True,
            keep_aspect_ratios: bool = True,
            header_cols: Optional[List[str]] = None,
            idx: Optional[Union[List[bool],List[int]]] = None,
            background_color: Tuple[int] = (0, 0, 0),
            **kwargs
            ) -> None:
        """Plots a grid of size (n_rows, n_cols) with images from the dataframe.

        Args:
            n_rows (int, optional): The number of rows in the grid.
            n_cols (int, optional): The number of columns in the grid.
            offset (float, optional): The offset between images.
            img_min (float, optional): The minimal size of the plotted images.
            rotate (bool, optional): Rotates the images to have the same orientation.
            keep_aspect_ratios (bool, optional): Whether aspect ratios are kept for images.
            header_cols (Optional[List[str]], optional): List of headers for each column.
            idx (Optional[Union[List[bool],List[int]]], optional): List of indices to plot. None plots random images. Index -1 plots an empty image.
            background_color (Tuple[int], optional): Background color of the grid.
        """

        if len(self.df) == 0:
            return None
        
        # Select indices of images to be plotted
        if idx is None:
            n = min(len(self.df), n_rows*n_cols)
            idx = np.random.permutation(len(self.df))[:n]
        else:
            if isinstance(idx, pd.Series):
                idx = idx.values
            if isinstance(idx[0], (bool, np.bool_)):
                idx = np.where(idx)[0]
            n = min(np.array(idx).size, n_rows*n_cols)
            idx = np.matrix.flatten(np.array(idx))[:n]

        # Load images and compute their ratio
        ratios = []
        ims = []
        for k in idx:
            if k >= 0:
                # Load the image with index k
                if self.load_label:
                    im, _ = self[k]
                else:
                    im = self[k]
                ims.append(im)
                ratios.append(im.size[0] / im.size[1])
            else:
                # Load a black image
                ims.append(Image.fromarray(np.zeros((2, 2), dtype = "uint8")))

        # Safeguard when all indices are -1
        if len(ratios) == 0:
            return None
        
        # Get the size of the images after being resized
        ratio = np.median(ratios)
        if ratio > 1:    
            img_w, img_h = int(img_min*ratio), int(img_min)
        else:
            img_w, img_h = int(img_min), int(img_min/ratio)

        # Compute height offset if headers are present
        if header_cols is not None:
            offset_h = 30
            if len(header_cols) != n_cols:
                raise(Exception("Length of header_cols must be the same as n_cols."))
        else:
            offset_h = 0

        # Create an empty image grid
        im_grid = Image.new('RGB', (n_cols*img_w + (n_cols-1)*offset, offset_h + n_rows*img_h + (n_rows-1)*offset), background_color)

        # Fill the grid image by image
        pos_y = offset_h
        for i in range(n_rows):
            row_h = 0
            for j in range(n_cols):
                k = (n_cols)*i + j
                if k < n:
                    # Possibly rotate the image
                    im = ims[k]
                    if rotate and ((ratio > 1 and im.size[0] < im.size[1]) or (ratio < 1 and im.size[0] > im.size[1])):
                        im = im.transpose(Image.Transpose.ROTATE_90)

                    # Rescale the image
                    if keep_aspect_ratios:
                        w, h = im.size
                        c = min(img_w / w, img_h / h)
                        im = im.resize((int(c*w), int(c*h)))
                    else:
                        im = im.resize((img_w,img_h))
                    row_h = max(row_h, im.size[1])

                    # Place the image on the grid
                    pos_x = j*img_w + j*offset
                    im_grid.paste(im, (pos_x,pos_y))
            if row_h > 0:
                pos_y += row_h + offset
        im_grid = im_grid.crop((0, 0, im_grid.size[0], pos_y-offset))
 
        # Plot the image and add column headers if present
        fig = plt.figure()
        fig.patch.set_visible(False)
        ax = fig.add_subplot(111)
        plt.axis('off')
        plt.imshow(im_grid)
        if header_cols is not None:
            color = kwargs.pop('color', 'white')
            ha = kwargs.pop('ha', 'center')
            va = kwargs.pop('va', 'center')
            for i, header in enumerate(header_cols):
                pos_x = (i+0.5)*img_w + i*offset
                pos_y = offset_h/2
                plt.text(pos_x, pos_y, str(header), color=color, ha=ha, va=va, **kwargs)
        return fig

# Alias for DatasetFactory
class DatasetFactory(WildlifeDataset):
    pass


import os
from copy import deepcopy
import numpy as np
import pandas as pd
from tqdm import tqdm
from .. import datasets, splits
from typing import Optional, Dict, List

def resize_dataset(
        dataset: datasets.WildlifeDataset,
        new_root: str,
        idx: Optional[List[int]] = None,
        copy_files: bool = True,
        remove_str: List[str] = [],
        replace_extensions: Dict[str, str] = {},
        ) -> pd.DataFrame:
    """Resizes dataset using `dataset.transform` into `new_root`.

    Args:
        dataset (datasets.WildlifeDataset): Dataset to be resized.
        new_root (str): Root to store new images.
        idx (Optional[List[int]], optional): If specified, then indices to consider.
        copy_files (bool, optional): Whether files should be copied as well or only datatframe created.
        remove_str (List[str], optional): List of strings to be removed from file names.
        replace_extensions (Dict[str, str], optional): Dictionary of extensions to be replaced.

    Returns:
        Description of the new dataset.
    """
    
    if idx is None:
        idx = range(len(dataset))
    
    df_new = []
    for i in tqdm(idx, mininterval=1, ncols=100):
        # Make image path unique across datasets
        row = dataset.metadata.iloc[i]
        base, ext = os.path.splitext(row[dataset.col_path])
        if ext in replace_extensions:
            ext = replace_extensions[ext]
        img_path = base + "_" + str(row["image_id"]) + ext
        img_path = img_path.replace('\\\\', '/')
        img_path = img_path.replace('\\', '/')
        for remove in remove_str:
            img_path = img_path.replace(remove, '')

        # Save image to new root with unique image path
        if copy_files:
            image = dataset[i]
            full_img_path = os.path.join(new_root, img_path)
            if not os.path.exists(os.path.dirname(full_img_path)):
                os.makedirs(os.path.dirname(full_img_path))
            image.save(full_img_path)

        # Extract species
        if 'species' in row:
            species = row['species']
        else:
            if hasattr(dataset, 'summary'):
                species = dataset.summary['animals']
                if len(species) != 1:
                    raise Exception('There should be only one species')
                species = list(species)[0]
            else:
                species = np.nan

        # Update dataframe
        df_new.append({
            'image_id': row['image_id'],
            dataset.col_label: row[dataset.col_label],
            dataset.col_path: img_path,
            'species': species,
            'date': row.get('date', np.nan),
            'orientation': row.get('orientation', np.nan),
        })
    return pd.DataFrame(df_new)

def get_every_k(
        dataset: datasets.WildlifeDataset,
        k: int,
        groupby_cols: str | List[str],
        ) -> List[int]:
    """Gets indices of every k-th image based on columns in `groupby_cols`.

    Args:
        dataset (datasets.WildlifeDataset): Dataset to be resized.
        k (int): Number of images to skip.
        groupby_cols (str | List[str]): For which groups the indices will be computed.

    Returns:
        Computed indices.
    """
    
    idx = np.array([], dtype=int)
    for _, df_red in dataset.df.groupby(groupby_cols):
        idx = np.hstack((idx, df_red.index[np.arange(0, len(df_red), k)]))
    # Convert loc to iloc
    idx = dataset.df.index.get_indexer(idx)
    return np.sort(idx)

def convert_dataset(dataset, new_root, k=1, col_label=None, **kwargs):
    if k > 1:
        if col_label is not None:
            idx = get_every_k(dataset, k, col_label)
        else:
            idx = get_every_k(dataset, k, dataset.col_label)        
        return resize_dataset(dataset, new_root, idx=idx, **kwargs)
    else:
        return resize_dataset(dataset, new_root, **kwargs)

def add_split_column(df, splitter):
    idx_train, idx_test = splitter.split(df)[0]
    df.loc[idx_train, 'split'] = 'train'
    df.loc[idx_test, 'split'] = 'test'
    return df

def prepare_aau_zebrafish(root, new_root, k=20, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.AAUZebraFish(root, img_load="bbox", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_aerial_cattle_2017(root, new_root, k=20, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.AerialCattle2017(root, img_load="full", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_amvrakikos_turtles(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.AmvrakikosTurtles(root, img_load="bbox", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_atrw(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.ATRW(root, img_load="bbox", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_beluga_id(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.BelugaIDv2(root, img_load="bbox", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.TimeProportionOpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_bird_individual_id(root, new_root, k=20, transform=None, add_split=False, splitter=None, segmented=True, **kwargs):
    if segmented:
        root = root + "Segmented"
    dataset = datasets.BirdIndividualIDSegmented(root, img_load="crop_black", transform=transform, remove_unknown=True)
    dataset = dataset.get_subset(~dataset.df['date'].isnull())
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.TimeProportionOpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_cat_individual_images(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.CatIndividualImages(root, img_load="full", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_chicks4free_id(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.Chicks4FreeID(root, img_load="full", transform=transform, remove_unknown=True)
    # Change the path from np.nan so that it is saved correctly
    dataset.df[dataset.col_path] = 'images/' + dataset.df['image_id'].astype('str') + '.jpg'
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_cow_dataset(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.CowDataset(root, img_load="full", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.TimeProportionOpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_cows2021(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.Cows2021v2(root, img_load="full", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.TimeProportionOpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_ctai(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.CTai(root, img_load="full", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_czoo(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.CZoo(root, img_load="full", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_dog_facenet(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.DogFaceNet(root, img_load="full", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_drosophila(root, new_root, k=1000, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.Drosophila(root, img_load="full", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_elpephants(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.ELPephants(root, img_load="full", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_friesian_cattle_2015(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.FriesianCattle2015v2(root, img_load="crop_black", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_friesian_cattle_2017(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.FriesianCattle2017(root, img_load="full", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_giraffes(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.Giraffes(root, img_load="full", transform=transform, remove_unknown=True)
    dataset = dataset.get_subset(~dataset.df['date'].isnull())
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.TimeProportionOpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_giraffe_zebra_id(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.GiraffeZebraID(root, img_load="bbox", transform=transform, remove_unknown=True)
    dataset = dataset.get_subset(~dataset.df['date'].isnull())
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.TimeProportionOpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_happy_whale(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.HappyWhale(root, img_load="full", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_humpback_whale_id(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.HumpbackWhaleID(root, img_load="full", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_hyena_id_2022(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.HyenaID2022(root, img_load="bbox", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_ipanda_50(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.IPanda50(root, img_load="full", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_leopard_id_2022(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.LeopardID2022(root, img_load="bbox", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_macaque_faces(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.MacaqueFaces(root, img_load="full", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_mpdd(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.MPDD(root, img_load="full", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_multi_cam_cows_2024(root, new_root, k=20, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.MultiCamCows2024(root, img_load="full", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.TimeProportionOpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_ndd20(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.NDD20v2(root, img_load="bbox_mask", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_noaa_right_whale(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.NOAARightWhale(root, img_load="full", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_nyala_data(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.NyalaData(root, img_load="full", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_open_cows_2020(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.OpenCows2020(root, img_load="full", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_polar_bear_vidid(root, new_root, k=100, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.PolarBearVidID(root, img_load="full", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_prim_face(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    kwargs = deepcopy(kwargs)
    remove_str = kwargs.pop('remove_str', [])
    remove_str.append(' ')
    kwargs['remove_str'] = remove_str
    dataset = datasets.PrimFace(root, img_load="crop_white", transform=transform, remove_unknown=True)    
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_reunion_turtles(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.ReunionTurtles(root, img_load="full", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_seal_id(root, new_root, k=1, transform=None, add_split=False, splitter=None, segmented=True, **kwargs):
    if segmented:
        root = root + "Segmented"
    dataset = datasets.SealIDSegmented(root, img_load="crop_black", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_sea_star_reid_2023(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.SeaStarReID2023(root, img_load="full", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.TimeProportionOpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_sea_turtle_id_2022(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.SeaTurtleID2022(root, img_load="bbox", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.TimeProportionOpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_sea_turtle_id_2022_animal_clef_2025(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.AnimalCLEF2025_SeaTurtleID2022(root, img_load="bbox", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_smalst(root, new_root, k=10, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.SMALST(root, img_load="bbox_mask", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_southern_province_turtles(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.SouthernProvinceTurtles(root, img_load="full", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_stripe_spotter(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.StripeSpotter(root, img_load="bbox", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_whaleshark_id(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.WhaleSharkID(root, img_load="bbox", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_zakynthos_turtles(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.ZakynthosTurtles(root, img_load="bbox", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

def prepare_zindi_turtle_recall(root, new_root, k=1, transform=None, add_split=False, splitter=None, **kwargs):
    dataset = datasets.ZindiTurtleRecall(root, img_load="full", transform=transform, remove_unknown=True)
    df = convert_dataset(dataset, new_root, k=k, **kwargs)
    if add_split:
        if splitter is None:
            splitter = splits.OpenSetSplit(0.8, 0.05, seed=666)
        return add_split_column(df, splitter)
    return df

# LionData are missing
prepare_functions = {
    'AAUZebraFish': prepare_aau_zebrafish,
    'AerialCattle2017': prepare_aerial_cattle_2017,
    'AmvrakikosTurtles': prepare_amvrakikos_turtles,
    'ATRW': prepare_atrw,
    'BelugaID': prepare_beluga_id,
    'BirdIndividualID': prepare_bird_individual_id,
    'CatIndividualImages': prepare_cat_individual_images,
    'Chicks4FreeID': prepare_chicks4free_id,
    'CowDataset': prepare_cow_dataset,
    'Cows2021': prepare_cows2021,
    'CTai': prepare_ctai,
    'CZoo': prepare_czoo,
    'DogFaceNet': prepare_dog_facenet,
    'Drosophila': prepare_drosophila,
    'ELPephants': prepare_elpephants,
    'FriesianCattle2015': prepare_friesian_cattle_2015,
    'FriesianCattle2017': prepare_friesian_cattle_2017,
    'Giraffes': prepare_giraffes,
    'GiraffeZebraID': prepare_giraffe_zebra_id,
    'HappyWhale': prepare_happy_whale,
    'HumpbackWhaleID': prepare_humpback_whale_id,
    'HyenaID2022': prepare_hyena_id_2022,
    'IPanda50': prepare_ipanda_50,
    'LeopardID2022': prepare_leopard_id_2022,
    'MacaqueFaces': prepare_macaque_faces,
    'MPDD': prepare_mpdd,
    'MultiCamCows2024': prepare_multi_cam_cows_2024,
    'NDD20': prepare_ndd20,
    'NOAARightWhale': prepare_noaa_right_whale,
    'NyalaData': prepare_nyala_data,
    'OpenCows2020': prepare_open_cows_2020,
    'PolarBearVidID': prepare_polar_bear_vidid,
    'PrimFace': prepare_prim_face,    
    'ReunionTurtles': prepare_reunion_turtles,
    'SealID': prepare_seal_id,
    'SeaStarReID2023': prepare_sea_star_reid_2023,
    'SeaTurtleID2022': prepare_sea_turtle_id_2022,
    'SeaTurtleID2022_AnimalCLEF2025': prepare_sea_turtle_id_2022_animal_clef_2025,
    'SMALST': prepare_smalst,
    'SouthernProvinceTurtles': prepare_southern_province_turtles,
    'StripeSpotter': prepare_stripe_spotter,
    'WhaleSharkID': prepare_whaleshark_id,
    'ZakynthosTurtles': prepare_zakynthos_turtles,
    'ZindiTurtleRecall': prepare_zindi_turtle_recall,
}

species_conversion = {
    'Anthenea australiae': 'sea star',
    'Asteria rubens': 'sea star',
    'bottlenose_dolphin': 'dolphin',
    'beluga': 'whale',
    'beluga_whale': 'whale',
    'blue_whale': 'whale',
    'BND': 'dolphin',
    'Friesian cattle': 'cow',
    'WBD': 'dolphin',
    'amur tiger': 'tiger',
    'beluga whale': 'whale',
    'brydes_whale': 'whale',
    'cat': 'cat',
    'chickens': 'chicken',
    'chimpanzee': 'chimpanzee',
    'commersons_dolphin': 'dolphin',
    'common_dolphin': 'dolphin',
    'cow': 'cow',
    'cuviers_beaked_whale': 'whale',
    'dog': 'dog',
    'drosophila': 'fly',
    'dusky_dolphin': 'dolphin',
    'elephant': 'elephant',
    'false_killer_whale': 'whale',
    'fin_whale': 'whale',
    'frasiers_dolphin': 'dolphin',
    'giraffe': 'giraffe',
    'giraffe_masai': 'giraffe',
    'globis': 'whale',
    'gray_whale': 'whale',
    'great panda': 'panda',
    'great_tits': 'bird',
    'green turtle': 'sea turtle',
    'Green': 'sea turtle',
    'Hawksbill': 'sea turtle',
    'hyena': 'hyena',
    'humpback_whale': 'whale',
    'japanese monkey': 'macaque',
    'killer_whale': 'whale',
    'leopard': 'leopard',
    'loggerhead turtle': 'sea turtle',
    'long_finned_pilot_whale': 'whale',
    'melon_headed_whale': 'whale',
    'minke_whale': 'whale',
    'nyala': 'nyala',
    'pantropic_spotted_dolphin': 'dolphin',
    'pilot_whale': 'whale',
    'polar bear': 'polar bear',
    'pygmy_killer_whale': 'whale',
    'right whale': 'whale',
    'ringed seal': 'seal',
    'rhesus macaque': 'macaque',
    'rhesus monkey': 'macaque',
    'rhincodon_typus': 'whaleshark',
    'rough_toothed_dolphin': 'dolphin',
    'sea turtle': 'sea turtle',
    'sei_whale': 'whale',
    'short_finned_pilot_whale': 'whale',
    'sociable_weavers': 'bird',
    'southern_right_whale': 'whale',
    'spinner_dolphin': 'dolphin',
    'spotted_dolphin': 'whale',
    'spotted hyena': 'hyena',
    'whale': 'whale',
    'whale shark': 'shark',
    'white_sided_dolphin': 'dolphin',
    'zebra': 'zebra',
    'zebra_finch': 'bird',
    'zebra_plains': 'zebra',
    'zebrafish': 'fish',
}



import os
import numpy as np
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class Giraffes(DatasetFactory):
    summary = summary['Giraffes']

    @classmethod
    def _download(cls):
        url = 'ftp://pbil.univ-lyon1.fr/pub/datasets/miele2021/'
        command = f"wget -rpk -l 10 -np -c --random-wait -U Mozilla {url} -P '.' "
        exception_text = '''Download works only on Linux. Please download it manually.
            Check https://wildlifedatasets.github.io/wildlife-datasets/downloads#giraffes'''
        if os.name == 'posix':
            os.system(command)
        else:
            raise Exception(exception_text)

    @classmethod
    def _extract(cls):
        pass

    def create_catalogue(self) -> pd.DataFrame:
        # Find all images in root
        data = utils.find_images(self.root)
        folders = data['path'].str.split(os.path.sep, expand=True)
        n_folders = max(folders.columns)

        # Extract information from the folder structure
        clusters = folders[n_folders-1] == 'clusters'
        data, folders = data[clusters], folders[clusters]

        # Finalize the dataframe
        df = pd.DataFrame({    
            'image_id': utils.create_id(data['file']),
            'path': data['path'] + os.path.sep + data['file'],
            'identity': folders[n_folders],
            'date': data['file'].apply(lambda x: self.extract_date(x))
        })
        return self.finalize_catalogue(df)
    
    def extract_date(self, x):
        date = x.split('_')[1]
        if date == 'None':
            return np.nan
        else:
            return f'{date[:4]}-{date[4:6]}-{date[6:]}'



import numpy as np
import pandas as pd
from typing import List, Tuple
from .balanced_split import BalancedSplit


class TimeAwareSplit(BalancedSplit):
    """Base class for `TimeProportionSplit` and `TimeCutoffSplit`.
    """

    def modify_df(self, df: pd.DataFrame) -> pd.DataFrame:
        """Prepares dataframe for splits.

        Removes identities specified in `self.identity_skip` (usually unknown identities).
        Convert the `date` column into a unified format.
        Add the `year` column.

        Args:
            df (pd.DataFrame): A dataframe of the data. It must contain columns `identity` and `date`.

        Returns:
            Modified dataframe of the data.
        """
        
        # Check if the DataFrame contain the column date.
        if 'date' not in df.columns:
            raise(Exception('Dataframe df does not contain column date.'))
        
        # Remove identities to be skipped
        df = df.copy()
        df = df[df[self.col_label] != self.identity_skip]

        # Removes entries without dates
        df = df[~df['date'].isnull()]
        
        # Convert date to datetime format (from possibly strings) and drop hours
        df['date'] = pd.to_datetime(df['date']).apply(lambda x: x.date())
        df['year'] = df['date'].apply(lambda x: x.year).to_numpy()            
        return df


class TimeProportionSplit(TimeAwareSplit):
    """Time-proportion non-random splitting method into training and testing sets.

    For each individual, it extracts unique observation dates
    and puts half to the training to the testing set.
    Ignores individuals with only one observation date.
    Implementation of [this paper](https://arxiv.org/abs/2211.10307).
    """

    def __init__(
            self,
            ratio: float = 0.5,
            **kwargs
            ):
        """Initializes the class.

        Args:
            ratio (float, optional): The fraction of dates going to the training set.
            **kwargs (type, optional): See kwargs `seed`, `identity_skip` and `col_label` of the parent class.
        """

        self.ratio = ratio
        super().__init__(**kwargs)

    def split(self, df: pd.DataFrame) -> List[Tuple[np.ndarray, np.ndarray]]:
        """Implementation of the [base splitting method](../reference_splits#splits.balanced_split.BalancedSplit.split).

        Args:
            df (pd.DataFrame): A dataframe of the data. It must contain columns `identity` and `date`.

        Returns:
            List of splits. Each split is list of labels of the training and testing sets.
        """
        
        df = self.modify_df(df)
        idx_train = []
        idx_test = []
        # Loop over all identities; x is a tuple (identity, df with unique identity)
        for _, df_name in df.groupby(self.col_label):            
            dates = df_name.groupby('date')
            n_dates = len(dates)
            if n_dates > 1:
                # Loop over all dates; y is a tuple (date, df with unique date and identity)
                for i, (_, df_date) in enumerate(dates):
                    # Add half dates to the training and half to the testing set
                    if i < np.minimum(n_dates-1, int(np.round(self.ratio*n_dates))):
                        idx_train += list(df_date.index)
                    else:
                        idx_test += list(df_date.index)
            else:
                idx_train += list(df_name.index)
        return [(np.array(idx_train), np.array(idx_test))]


class TimeProportionOpenSetSplit(TimeAwareSplit):
    """Time-proportion open set splitting method into training and testing sets.

    First, it pust some individuals into the training set only.
    Then it is the TimeProportionSplit.
    """

    def __init__(
            self,
            ratio_train: float,
            ratio_class_test: float = None,
            n_class_test: int = None,
            **kwargs
            ) -> None:
        """Initializes the class.

        The user must provide exactly one from `ratio_class_test` and `n_class_test`.
        The latter specifies the number of individuals to be only in the testing set.
        The former specified the ratio of samples of individuals (not individuals themselves)
        to be only in the testing set.

        Args:
            ratio_train (float): *Approximate* size of the training set.
            ratio_class_test (float, optional): *Approximate* ratio of samples of individuals only in the testing set.
            n_class_test (int, optional): Number of individuals only in the testing set.
            seed (int, optional): Initial seed for the LCG random generator.
            identity_skip (str, optional): Name of the identities to ignore.
            col_label (str, optional): Column name containing individual animal names (labels).
        """

        if ratio_class_test is None and n_class_test is None:
            raise(Exception('Either ratio_class_test or n_class_test must be provided.'))
        elif ratio_class_test is not None and n_class_test is not None:
            raise(Exception('Only ratio_class_test or n_class_test can be provided.'))
        
        self.ratio_train = ratio_train
        self.ratio_class_test = ratio_class_test
        self.n_class_test = n_class_test
        super().__init__(**kwargs)

    def split(self, df: pd.DataFrame) -> List[Tuple[np.ndarray, np.ndarray]]:
        """Implementation of the [base splitting method](../reference_splits#splits.balanced_split.BalancedSplit.split).

        Args:
            df (pd.DataFrame): A dataframe of the data. It must contain columns `identity` and `date`.

        Returns:
            List of splits. Each split is list of labels of the training and testing sets.
        """
        
        df = self.modify_df(df)

        # Initialize the random number generator
        lcg = self.initialize_lcg()

        # Compute the counts and randomly permute them
        y_counts = df[self.col_label].value_counts()
        n_class = len(y_counts)
        idx = lcg.random_permutation(n_class)
        y_counts = y_counts.iloc[idx]

        # Compute number of identities in the testing set
        n = len(df)
        if self.n_class_test is None:
            n_test = np.round(n*self.ratio_class_test).astype(int)
            n_class_test = np.where(np.cumsum(y_counts) >= n_test)[0][0] + 1
        else:
            n_class_test = self.n_class_test

        # Specify individuals going purely into training and testing sets
        individual_train = np.array([], dtype=object)
        individual_test = np.array(y_counts.index[:n_class_test])

        # Compute how many samples go automatically to the training and testing sets
        y_counts = df[self.col_label].value_counts()
        n_train = sum([y_counts.loc[y] for y in individual_train])
        n_test = sum([y_counts.loc[y] for y in individual_test])
        
        # Recompute ratio_train and adjust it to proper bounds
        ratio_train = self.ratio_train
        if n_train + n_test > 0 and n_train + n_test < n:
            ratio_train = (n*ratio_train - n_train) / (n - n_test - n_train)
        ratio_train = np.clip(ratio_train, 0, 1)

        idx_train = []
        idx_test = []
        # Loop over all identities; x is a tuple (identity, df with unique identity)
        for name, df_name in df.groupby(self.col_label):
            if name in individual_train and name in individual_test:
                # Check if the class does not belong to both sets
                raise(Exception('Individual cannot be both in individual_train and individual_test.'))
            elif name in individual_train:
                # Check if the class does not belong to the training set
                idx_train += list(df_name.index)
            elif name in individual_test:
                # Check if the class does not belong to the testing set
                idx_test += list(df_name.index)
            else:
                dates = df_name.groupby('date')
                n_dates = len(dates)
                if n_dates > 1:
                    # Loop over all dates; y is a tuple (date, df with unique date and identity)
                    for i, (_, df_date) in enumerate(dates):
                        # Add half dates to the training and half to the testing set
                        if i < np.minimum(n_dates-1, int(np.round(ratio_train*n_dates))):
                            idx_train += list(df_date.index)
                        else:
                            idx_test += list(df_date.index)
                else:
                    idx_train += list(df_name.index)
        return [(np.array(idx_train), np.array(idx_test))]


class TimeCutoffSplit(TimeAwareSplit):
    """Time-cutoff non-random splitting method into training and testing sets.

    Puts all individuals observed before `year` into the training test.
    Puts all individuals observed during `year` into the testing test.
    Ignores all individuals observed after `year`.
    Implementation of [this paper](https://arxiv.org/abs/2211.10307).
    """

    def __init__(
            self,
            year: int,
            test_one_year_only: bool = True,
            **kwargs
            ) -> None:
        """Initializes the class.

        Args:
            year (int): Splitting year.
            test_one_year_only (bool, optional): Whether the test set is `df['year'] == year` or `df['year'] >= year`.
            **kwargs (type, optional): See kwargs `seed`, `identity_skip` and `col_label` of the parent class.
        """

        self.year = year
        self.test_one_year_only = test_one_year_only
        super().__init__(**kwargs)
    
    def split(self, df: pd.DataFrame) -> List[Tuple[np.ndarray, np.ndarray]]:
        """Implementation of the [base splitting method](../reference_splits#splits.balanced_split.BalancedSplit.split).

        Args:
            df (pd.DataFrame): A dataframe of the data. It must contain columns `identity` and `date`.

        Returns:
            List of labels of the training and testing sets.
        """

        df = self.modify_df(df)
        idx_train = list(np.where(df['year'] < self.year)[0])
        if self.test_one_year_only:
            idx_test = list(np.where(df['year'] == self.year)[0])
        else:
            idx_test = list(np.where(df['year'] >= self.year)[0])
        return [(np.array(df.index.values)[idx_train], np.array(df.index.values)[idx_test])]


class TimeCutoffSplitAll(TimeAwareSplit):
    """Sequence of time-cutoff splits TimeCutoffSplit for all possible years.

    Puts all individuals observed before `year` into the training test.
    Puts all individuals observed during `year` into the testing test.
    Implementation of [this paper](https://arxiv.org/abs/2211.10307).
    """

    def __init__(
            self,
            test_one_year_only: bool = True,
            **kwargs
            ) -> None:
        """Initializes the class.

        Args:
            test_one_year_only (bool, optional): Whether the test set is `df['year'] == year` or `df['year'] >= year`.
            **kwargs (type, optional): See kwargs `seed`, `identity_skip` and `col_label` of the parent class.
        """

        self.test_one_year_only = test_one_year_only
        super().__init__(**kwargs)
    
    def split(self, df: pd.DataFrame) -> List[Tuple[np.ndarray, np.ndarray]]:
        """Implementation of the [base splitting method](../reference_splits#splits.balanced_split.BalancedSplit.split).

        Args:
            df (pd.DataFrame): A dataframe of the data. It must contain columns `identity` and `date`.

        Returns:
            List of splits. Each split is list of labels of the training and testing sets.
        """

        df = self.modify_df(df)
        years = np.sort(df['year'].unique())[1:]
        splits = []
        for year in years:
            splitter = TimeCutoffSplit(
                year,
                test_one_year_only=self.test_one_year_only,
                seed=self.seed,
                identity_skip=self.identity_skip,
                col_label=self.col_label
                )
            for split in splitter.split(df):
                splits.append(split)
        return splits


class RandomProportion():
    """Wrapper for resplits of TimeProportionSplit.
    """

    def __init__(self, **kwargs):
        self.splitter = TimeProportionSplit(**kwargs)
    
    def split(self, df):
        splits = []
        for idx_train, idx_test in self.splitter.split(df):
            splits.append(self.splitter.resplit_random(df, idx_train, idx_test))
        return splits
    
    def set_col_label(self, *args, **kwargs):
        self.splitter.set_col_label(*args, **kwargs)



import os
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class ZakynthosTurtles(DatasetFactory):    
    summary = summary['ZakynthosTurtles']
    archive = 'zakynthosturtles.zip'

    @classmethod
    def _download(cls):
        command = f"datasets download -d wildlifedatasets/zakynthosturtles --force"
        exception_text = '''Kaggle must be setup.
            Check https://wildlifedatasets.github.io/wildlife-datasets/preprocessing#zakynthosturtles'''
        utils.kaggle_download(command, exception_text=exception_text, required_file=cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)

    def create_catalogue(self) -> pd.DataFrame:
        data = pd.read_csv(os.path.join(self.root, 'annotations.csv'))
        bbox = pd.read_csv(os.path.join(self.root, 'bbox.csv'))
        data = pd.merge(data, bbox, left_on='path', right_on='image_name')

        dates = data['date'].str.split('_')
        dates = dates.apply(lambda x: x[2] + '-' + x[1] + '-' + x[0])
        df = pd.DataFrame({
            'image_id': range(len(data)),
            'path': 'images/' + data['path'],
            'identity': data['identity'],
            'date': dates,
            'orientation': data['orientation'],
            'bbox': data[['bbox_x', 'bbox_y', 'bbox_width', 'bbox_height']].values.tolist()
        })
        return self.finalize_catalogue(df)



import os
import numpy as np
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class ReunionTurtles(DatasetFactory):    
    summary = summary['ReunionTurtles']
    archive = 'reunionturtles.zip'

    @classmethod
    def _download(cls):
        command = f"datasets download -d wildlifedatasets/reunionturtles --force"
        exception_text = '''Kaggle must be setup.
            Check https://wildlifedatasets.github.io/wildlife-datasets/preprocessing#reunionturtles'''
        utils.kaggle_download(command, exception_text=exception_text, required_file=cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)

    def create_catalogue(self) -> pd.DataFrame:
        data = pd.read_csv(os.path.join(self.root, 'data.csv'))

        date = pd.to_datetime(data['Date'])
        year = date.apply(lambda x: x.year)
        path = data['Species'] + os.path.sep + data['Turtle_ID'] + os.path.sep + year.astype(str) + os.path.sep + data['Photo_name']
        orientation = data['Photo_name'].apply(lambda x: os.path.splitext(x)[0].split('_')[2])
        orientation = orientation.replace({'L': 'left', 'R': 'right'})

        # Extract and convert ID codes
        id_code = list(data['ID_Code'].apply(lambda x: x.split(';')))
        max0 = 0
        max1 = 0
        for x in id_code:
            for y in x:
                max0 = max(max0, int(y[0]))
                max1 = max(max1, int(y[1]))
        code = np.zeros((len(id_code), max0, max1), dtype=int)
        for i, x in enumerate(id_code):
            for y in x:
                code[i, int(y[0])-1, int(y[1])-1] = int(y[2])
        code = code.reshape(len(id_code), -1)

        # Finalize the dataframe
        df = pd.DataFrame({
            'image_id': range(len(data)),
            'path': path,
            'identity': data['Turtle_ID'],
            'date': date,
            'orientation': orientation,
            'species': data['Species'],
            'id_code': list(code)
        })
        return self.finalize_catalogue(df)



from .analysis import analyze_split, extract_data_split, recognize_id_split, recognize_time_split, visualize_split
from .balanced_split import BalancedSplit
from .identity_split import FullSplit, IdentitySplit, ClosedSetSplit, OpenSetSplit, DisjointSetSplit
from .lcg import Lcg
from .time_aware_split import TimeAwareSplit, TimeProportionSplit, TimeProportionOpenSetSplit, TimeCutoffSplit, TimeCutoffSplitAll, RandomProportion



import os
import shutil
import numpy as np
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class NyalaData(DatasetFactory):
    summary = summary['NyalaData']
    url = 'https://github.com/tvanzyl/wildlife_reidentification/archive/refs/heads/main.zip'
    archive = 'main.zip'

    @classmethod
    def _download(cls):
        utils.download_url(cls.url, cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)
        shutil.rmtree('wildlife_reidentification-main/Lion_Data_Zero')

    def create_catalogue(self) -> pd.DataFrame:
        # Find all images in root
        data = utils.find_images(self.root)
        folders = data['path'].str.split(os.path.sep, expand=True)

        # Extract information from the folder structure and about orientation
        identity = folders[3].astype(int)
        orientation = np.full(len(data), np.nan, dtype=object)
        orientation[data['file'].str.contains('left')] = 'left'
        orientation[data['file'].str.contains('right')] = 'right'

        # Finalize the dataframe
        df = pd.DataFrame({
            'image_id': utils.create_id(data['file']),
            'path': data['path'] + os.path.sep + data['file'],
            'identity': identity,
            'orientation': orientation,
            'original_split': folders[2]
        })
        return self.finalize_catalogue(df)   



import numpy as np
import pandas as pd
from .datasets import DatasetFactory
from .summary import summary
from datasets import load_dataset

class Chicks4FreeID(DatasetFactory):
    summary = summary['Chicks4FreeID']
    hf_url = 'dariakern/Chicks4FreeID'
    determined_by_df = False
    saved_to_system_folder = True

    @classmethod
    def _download(cls, hf_option = 'chicken-re-id-all-visibility'):
        load_dataset(cls.hf_url, hf_option)

    @classmethod
    def _extract(cls, **kwargs):
        pass

    def create_catalogue(self, hf_option = 'chicken-re-id-all-visibility') -> pd.DataFrame:
        dataset = load_dataset(self.hf_url, hf_option)
        
        self.n_train = dataset['train'].num_rows
        self.n_test = dataset['test'].num_rows
        self.dataset = dataset
        return pd.DataFrame({
            'image_id': range(self.n_train+self.n_test),
            'identity': dataset['train']['identity'] + dataset['test']['identity'],
            'path': np.nan,
            'split_original': self.n_train*['train'] + self.n_test*['test']
        })

    def get_image(self, idx):
        if idx < self.n_train:
            return self.dataset['train'][int(idx)]['crop']
        else:
            return self.dataset['test'][int(idx)-self.n_train]['crop']
